@@ -1,3 +1,12 @@
+#include "server.h"
+#include "log.h"
+#include "stream.h"
+#include "plugin.h"
+
+#include "configparser.h"
+#include "configfile.h"
+#include "proc_open.h"
+
 #include <sys/stat.h>
 
 #include <stdlib.h>
@@ -10,18 +19,6 @@
 #include <limits.h>
 #include <assert.h>
 
-#include "server.h"
-#include "log.h"
-#include "stream.h"
-#include "plugin.h"
-#ifdef USE_LICENSE
-#include "license.h"
-#endif
-
-#include "configparser.h"
-#include "configfile.h"
-#include "proc_open.h"
-
 
 static int config_insert(server *srv) {
 	size_t i;
@@ -42,12 +39,12 @@ static int config_insert(server *srv) {
 
 		{ "server.event-handler",        NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_SERVER },      /* 10 */
 		{ "server.pid-file",             NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_SERVER },      /* 11 */
-		{ "server.max-request-size",     NULL, T_CONFIG_SHORT, T_CONFIG_SCOPE_CONNECTION },   /* 12 */
+		{ "server.max-request-size",     NULL, T_CONFIG_INT, T_CONFIG_SCOPE_CONNECTION },     /* 12 */
 		{ "server.max-worker",           NULL, T_CONFIG_SHORT, T_CONFIG_SCOPE_SERVER },       /* 13 */
 		{ "server.document-root",        NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },  /* 14 */
-		{ "server.force-lowercase-filenames", NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },   /* 15 */
+		{ "server.force-lowercase-filenames", NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },/* 15 */
 		{ "debug.log-condition-handling", NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },    /* 16 */
-		{ "server.max-keep-alive-requests", NULL, T_CONFIG_SHORT, T_CONFIG_SCOPE_CONNECTION }, /* 17 */
+		{ "server.max-keep-alive-requests", NULL, T_CONFIG_SHORT, T_CONFIG_SCOPE_CONNECTION },/* 17 */
 		{ "server.name",                 NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_CONNECTION },  /* 18 */
 		{ "server.max-keep-alive-idle",  NULL, T_CONFIG_SHORT, T_CONFIG_SCOPE_CONNECTION },   /* 19 */
 
@@ -91,9 +88,26 @@ static int config_insert(server *srv) {
 		{ "server.core-files",           NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 46 */
 		{ "ssl.cipher-list",             NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_SERVER },      /* 47 */
 		{ "ssl.use-sslv2",               NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 48 */
-		{ "etag.use-inode",             NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER }, /* 49 */
-		{ "etag.use-mtime",             NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER }, /* 50 */
-		{ "etag.use-size",             NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER }, /* 51 */
+		{ "etag.use-inode",              NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },     /* 49 */
+		{ "etag.use-mtime",              NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },     /* 50 */
+		{ "etag.use-size",               NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },     /* 51 */
+		{ "server.reject-expect-100-with-417",  NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER }, /* 52 */
+		{ "debug.log-timeouts",          NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 53 */
+		{ "server.defer-accept",         NULL, T_CONFIG_SHORT, T_CONFIG_SCOPE_CONNECTION },   /* 54 */
+		{ "server.breakagelog",          NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_SERVER },      /* 55 */
+		{ "ssl.verifyclient.activate",   NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },     /* 56 */
+		{ "ssl.verifyclient.enforce",    NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },     /* 57 */
+		{ "ssl.verifyclient.depth",      NULL, T_CONFIG_SHORT,   T_CONFIG_SCOPE_SERVER },     /* 58 */
+		{ "ssl.verifyclient.username",   NULL, T_CONFIG_STRING,  T_CONFIG_SCOPE_SERVER },     /* 59 */
+		{ "ssl.verifyclient.exportcert", NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },     /* 60 */
+
+		{ "server.set-v6only",           NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION }, /* 61 */
+		{ "ssl.use-sslv3",               NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },     /* 62 */
+		{ "ssl.dh-file",                 NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_SERVER },      /* 63 */
+		{ "ssl.ec-curve",                NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_SERVER },      /* 64 */
+		{ "ssl.disable-client-renegotiation", NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },/* 65 */
+		{ "ssl.honor-cipher-order",      NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_SERVER },     /* 66 */
+
 		{ "server.host",                 "use server.bind instead", T_CONFIG_DEPRECATED, T_CONFIG_SCOPE_UNSET },
 		{ "server.docroot",              "use server.document-root instead", T_CONFIG_DEPRECATED, T_CONFIG_SCOPE_UNSET },
 		{ "server.virtual-root",         "load mod_simple_vhost and use simple-vhost.server-root instead", T_CONFIG_DEPRECATED, T_CONFIG_SCOPE_UNSET },
@@ -135,6 +149,9 @@ static int config_insert(server *srv) {
 
 	cv[43].destination = &(srv->srvconf.max_conns);
 	cv[12].destination = &(srv->srvconf.max_request_size);
+	cv[52].destination = &(srv->srvconf.reject_expect_100_with_417);
+	cv[55].destination = srv->srvconf.breakagelog_file;
+
 	srv->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));
 
 	assert(srv->config_storage);
@@ -152,6 +169,8 @@ static int config_insert(server *srv) {
 		s->error_handler = buffer_init();
 		s->server_tag    = buffer_init();
 		s->ssl_cipher_list = buffer_init();
+		s->ssl_dh_file   = buffer_init();
+		s->ssl_ec_curve  = buffer_init();
 		s->errorfile_prefix = buffer_init();
 		s->max_keep_alive_requests = 16;
 		s->max_keep_alive_idle = 5;
@@ -159,8 +178,12 @@ static int config_insert(server *srv) {
 		s->max_write_idle = 360;
 		s->use_xattr     = 0;
 		s->is_ssl        = 0;
-		s->ssl_use_sslv2 = 1;
+		s->ssl_honor_cipher_order = 1;
+		s->ssl_use_sslv2 = 0;
+		s->ssl_use_sslv3 = 1;
 		s->use_ipv6      = 0;
+		s->set_v6only    = 1;
+		s->defer_accept  = 0;
 #ifdef HAVE_LSTAT
 		s->follow_symlink = 1;
 #endif
@@ -170,15 +193,23 @@ static int config_insert(server *srv) {
 		s->etag_use_mtime = 1;
 		s->etag_use_size  = 1;
 		s->range_requests = 1;
-		s->force_lowercase_filenames = 0;
+		s->force_lowercase_filenames = (i == 0) ? 2 : 0; /* we wan't to detect later if user changed this for global section */
 		s->global_kbytes_per_second = 0;
 		s->global_bytes_per_second_cnt = 0;
 		s->global_bytes_per_second_cnt_ptr = &s->global_bytes_per_second_cnt;
+		s->ssl_verifyclient = 0;
+		s->ssl_verifyclient_enforce = 1;
+		s->ssl_verifyclient_username = buffer_init();
+		s->ssl_verifyclient_depth = 9;
+		s->ssl_verifyclient_export_cert = 0;
+		s->ssl_disable_client_renegotiation = 1;
 
 		cv[2].destination = s->errorfile_prefix;
 
 		cv[7].destination = s->server_tag;
 		cv[8].destination = &(s->use_ipv6);
+		cv[61].destination = &(s->set_v6only);
+		cv[54].destination = &(s->defer_accept);
 
 
 		/* 13 max-worker */
@@ -207,6 +238,7 @@ static int config_insert(server *srv) {
 		cv[33].destination = &(s->log_response_header);
 		cv[34].destination = &(s->log_request_header);
 		cv[35].destination = &(s->log_ssl_noise);
+		cv[53].destination = &(s->log_timeouts);
 
 		cv[36].destination = &(s->allow_http11);
 		cv[39].destination = s->ssl_ca_file;
@@ -214,10 +246,23 @@ static int config_insert(server *srv) {
 
 		cv[47].destination = s->ssl_cipher_list;
 		cv[48].destination = &(s->ssl_use_sslv2);
+		cv[62].destination = &(s->ssl_use_sslv3);
+		cv[63].destination = s->ssl_dh_file;
+		cv[64].destination = s->ssl_ec_curve;
+		cv[66].destination = &(s->ssl_honor_cipher_order);
+
 		cv[49].destination = &(s->etag_use_inode);
 		cv[50].destination = &(s->etag_use_mtime);
 		cv[51].destination = &(s->etag_use_size);
 
+		/* ssl.verify */
+		cv[56].destination = &(s->ssl_verifyclient);
+		cv[57].destination = &(s->ssl_verifyclient_enforce);
+		cv[58].destination = &(s->ssl_verifyclient_depth);
+		cv[59].destination = s->ssl_verifyclient_username;
+		cv[60].destination = &(s->ssl_verifyclient_export_cert);
+		cv[65].destination = &(s->ssl_disable_client_renegotiation);
+
 		srv->config_storage[i] = s;
 
 		if (0 != (ret = config_insert_values_global(srv, ((data_config *)srv->config_context->data[i])->value, cv))) {
@@ -283,19 +328,34 @@ int config_setup_connection(server *srv, connection *con) {
 	PATCH(log_condition_handling);
 	PATCH(log_file_not_found);
 	PATCH(log_ssl_noise);
+	PATCH(log_timeouts);
 
 	PATCH(range_requests);
 	PATCH(force_lowercase_filenames);
 	PATCH(is_ssl);
 
 	PATCH(ssl_pemfile);
+#ifdef USE_OPENSSL
+	PATCH(ssl_ctx);
+#endif
 	PATCH(ssl_ca_file);
 	PATCH(ssl_cipher_list);
+	PATCH(ssl_dh_file);
+	PATCH(ssl_ec_curve);
+	PATCH(ssl_honor_cipher_order);
 	PATCH(ssl_use_sslv2);
+	PATCH(ssl_use_sslv3);
 	PATCH(etag_use_inode);
 	PATCH(etag_use_mtime);
 	PATCH(etag_use_size);
- 
+
+	PATCH(ssl_verifyclient);
+	PATCH(ssl_verifyclient_enforce);
+	PATCH(ssl_verifyclient_depth);
+	PATCH(ssl_verifyclient_username);
+	PATCH(ssl_verifyclient_export_cert);
+	PATCH(ssl_disable_client_renegotiation);
+
 	return 0;
 }
 
@@ -344,14 +404,25 @@ int config_patch_connection(server *srv, connection *con, comp_key_t comp) {
 				PATCH(etag_use_size);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.pemfile"))) {
 				PATCH(ssl_pemfile);
+#ifdef USE_OPENSSL
+				PATCH(ssl_ctx);
+#endif
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.ca-file"))) {
 				PATCH(ssl_ca_file);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.honor-cipher-order"))) {
+				PATCH(ssl_honor_cipher_order);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.use-sslv2"))) {
 				PATCH(ssl_use_sslv2);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.use-sslv3"))) {
+				PATCH(ssl_use_sslv3);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.cipher-list"))) {
 				PATCH(ssl_cipher_list);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.engine"))) {
 				PATCH(is_ssl);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.dh-file"))) {
+				PATCH(ssl_dh_file);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.ec-curve"))) {
+				PATCH(ssl_ec_curve);
 #ifdef HAVE_LSTAT
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("server.follow-symlink"))) {
 				PATCH(follow_symlink);
@@ -374,6 +445,8 @@ int config_patch_connection(server *srv, connection *con, comp_key_t comp) {
 				PATCH(log_file_not_found);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("debug.log-ssl-noise"))) {
 				PATCH(log_ssl_noise);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("debug.log-timeouts"))) {
+				PATCH(log_timeouts);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("server.protocol-http11"))) {
 				PATCH(allow_http11);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("server.force-lowercase-filenames"))) {
@@ -382,6 +455,18 @@ int config_patch_connection(server *srv, connection *con, comp_key_t comp) {
 				PATCH(global_kbytes_per_second);
 				PATCH(global_bytes_per_second_cnt);
 				con->conf.global_bytes_per_second_cnt_ptr = &s->global_bytes_per_second_cnt;
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.verifyclient.activate"))) {
+				PATCH(ssl_verifyclient);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.verifyclient.enforce"))) {
+				PATCH(ssl_verifyclient_enforce);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.verifyclient.depth"))) {
+				PATCH(ssl_verifyclient_depth);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.verifyclient.username"))) {
+				PATCH(ssl_verifyclient_username);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.verifyclient.exportcert"))) {
+				PATCH(ssl_verifyclient_export_cert);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("ssl.disable-client-renegotiation"))) {
+				PATCH(ssl_disable_client_renegotiation);
 			}
 		}
 	}
@@ -915,7 +1000,7 @@ int config_parse_file(server *srv, config_t *context, const char *fn) {
 	return ret;
 }
 
-static char* getCWD() {
+static char* getCWD(void) {
 	char *s, *s1;
 	size_t len;
 #ifdef PATH_MAX
@@ -940,7 +1025,6 @@ static char* getCWD() {
 }
 
 int config_parse_cmd(server *srv, config_t *context, const char *cmd) {
-	proc_handler_t proc;
 	tokenizer_t t;
 	int ret;
 	buffer *source;
@@ -960,7 +1044,7 @@ int config_parse_cmd(server *srv, config_t *context, const char *cmd) {
 		chdir(context->basedir->ptr);
 	}
 
-	if (0 != proc_open_buffer(&proc, cmd, NULL, out, NULL)) {
+	if (0 != proc_open_buffer(cmd, NULL, out, NULL)) {
 		log_error_write(srv, __FILE__, __LINE__, "sbss",
 				"opening", source, "failed:", strerror(errno));
 		ret = -1;
@@ -1001,13 +1085,11 @@ int config_read(server *srv, const char *fn) {
 	context_init(srv, &context);
 	context.all_configs = srv->config_context;
 
-	pos = strrchr(fn,
 #ifdef __WIN32
-			'\\'
+	pos = strrchr(fn, '\\');
 #else
-			'/'
+	pos = strrchr(fn, '/');
 #endif
-			);
 	if (pos) {
 		buffer_copy_string_len(context.basedir, fn, pos - fn + 1);
 		fn = pos + 1;
@@ -1125,25 +1207,27 @@ int config_set_defaults(server *srv) {
 
 	struct ev_map { fdevent_handler_t et; const char *name; } event_handlers[] =
 	{
-		/* - poll is most reliable
+		/* - epoll is most reliable
 		 * - select works everywhere
-		 * - linux-* are experimental
 		 */
+#ifdef USE_LINUX_EPOLL
+		{ FDEVENT_HANDLER_LINUX_SYSEPOLL, "linux-sysepoll" },
+#endif
 #ifdef USE_POLL
 		{ FDEVENT_HANDLER_POLL,           "poll" },
 #endif
 #ifdef USE_SELECT
 		{ FDEVENT_HANDLER_SELECT,         "select" },
 #endif
-#ifdef USE_LINUX_EPOLL
-		{ FDEVENT_HANDLER_LINUX_SYSEPOLL, "linux-sysepoll" },
-#endif
-#ifdef USE_LINUX_SIGIO
-		{ FDEVENT_HANDLER_LINUX_RTSIG,    "linux-rtsig" },
+#ifdef USE_LIBEV
+		{ FDEVENT_HANDLER_LIBEV,          "libev" },
 #endif
 #ifdef USE_SOLARIS_DEVPOLL
 		{ FDEVENT_HANDLER_SOLARIS_DEVPOLL,"solaris-devpoll" },
 #endif
+#ifdef USE_SOLARIS_PORT
+		{ FDEVENT_HANDLER_SOLARIS_PORT,   "solaris-eventports" },
+#endif
 #ifdef USE_FREEBSD_KQUEUE
 		{ FDEVENT_HANDLER_FREEBSD_KQUEUE, "freebsd-kqueue" },
 		{ FDEVENT_HANDLER_FREEBSD_KQUEUE, "kqueue" },
@@ -1184,35 +1268,39 @@ int config_set_defaults(server *srv) {
 
 	buffer_to_lower(srv->tmp_buf);
 
-	if (0 == stat(srv->tmp_buf->ptr, &st1)) {
-		int is_lower = 0;
+	if (2 == s->force_lowercase_filenames) { /* user didn't configure it in global section? */
+		s->force_lowercase_filenames = 0; /* default to 0 */
 
-		is_lower = buffer_is_equal(srv->tmp_buf, s->document_root);
+		if (0 == stat(srv->tmp_buf->ptr, &st1)) {
+			int is_lower = 0;
 
-		/* lower-case existed, check upper-case */
-		buffer_copy_string_buffer(srv->tmp_buf, s->document_root);
+			is_lower = buffer_is_equal(srv->tmp_buf, s->document_root);
 
-		buffer_to_upper(srv->tmp_buf);
+			/* lower-case existed, check upper-case */
+			buffer_copy_string_buffer(srv->tmp_buf, s->document_root);
 
-		/* we have to handle the special case that upper and lower-casing results in the same filename
-		 * as in server.document-root = "/" or "/12345/" */
+			buffer_to_upper(srv->tmp_buf);
 
-		if (is_lower && buffer_is_equal(srv->tmp_buf, s->document_root)) {
-			/* lower-casing and upper-casing didn't result in
-			 * an other filename, no need to stat(),
-			 * just assume it is case-sensitive. */
+			/* we have to handle the special case that upper and lower-casing results in the same filename
+			 * as in server.document-root = "/" or "/12345/" */
 
-			s->force_lowercase_filenames = 0;
-		} else if (0 == stat(srv->tmp_buf->ptr, &st2)) {
+			if (is_lower && buffer_is_equal(srv->tmp_buf, s->document_root)) {
+				/* lower-casing and upper-casing didn't result in
+				 * an other filename, no need to stat(),
+				 * just assume it is case-sensitive. */
 
-			/* upper case exists too, doesn't the FS handle this ? */
+				s->force_lowercase_filenames = 0;
+			} else if (0 == stat(srv->tmp_buf->ptr, &st2)) {
 
-			/* upper and lower have the same inode -> case-insensitve FS */
+				/* upper case exists too, doesn't the FS handle this ? */
 
-			if (st1.st_ino == st2.st_ino) {
 				/* upper and lower have the same inode -> case-insensitve FS */
 
-				s->force_lowercase_filenames = 1;
+				if (st1.st_ino == st2.st_ino) {
+					/* upper and lower have the same inode -> case-insensitve FS */
+
+					s->force_lowercase_filenames = 1;
+				}
 			}
 		}
 	}
