@@ -420,6 +420,7 @@ static int connection_handle_write_prepare(server *srv, connection *con) {
 		case HTTP_METHOD_POST:
 		case HTTP_METHOD_HEAD:
 		case HTTP_METHOD_PUT:
+		case HTTP_METHOD_PATCH:
 		case HTTP_METHOD_MKCOL:
 		case HTTP_METHOD_DELETE:
 		case HTTP_METHOD_COPY:
@@ -1360,9 +1361,7 @@ connection *connection_accept(server *srv, server_socket *srv_socket) {
 			}
 
 			con->renegotiations = 0;
-#ifndef OPENSSL_NO_TLSEXT
 			SSL_set_app_data(con->ssl, con);
-#endif
 			SSL_set_accept_state(con->ssl);
 			con->conf.is_ssl=1;
 
@@ -1714,11 +1713,18 @@ int connection_state_machine(server *srv, connection *con) {
 										ERR_error_string(err, NULL));
 							} while((err = ERR_get_error()));
 						} else if (errno != 0) { /* ssl bug (see lighttpd ticket #2213): sometimes errno == 0 */
-							log_error_write(srv, __FILE__, __LINE__, "sddds", "SSL (error):",
+							switch(errno) {
+							case EPIPE:
+							case ECONNRESET:
+								break;
+							default:
+								log_error_write(srv, __FILE__, __LINE__, "sddds", "SSL (error):",
 									ssl_r, ret, errno,
 									strerror(errno));
+								break;
+							}
 						}
-	
+
 						break;
 					default:
 						while((err = ERR_get_error())) {
@@ -1726,7 +1732,7 @@ int connection_state_machine(server *srv, connection *con) {
 									ssl_r, ret,
 									ERR_error_string(err, NULL));
 						}
-	
+
 						break;
 					}
 				}
