@@ -241,14 +241,14 @@ static array *extract_forward_array(buffer *pbuffer)
 		int in_str = 0;
 		for (base = pbuffer->ptr, curr = pbuffer->ptr; *curr; curr++) {
 			if (in_str) {
-				if ((*curr > '9' || *curr < '0') && *curr != '.' && *curr != ':') {
+				if ((*curr > '9' || *curr < '0') && *curr != '.' && *curr != ':' && (*curr < 'a' || *curr > 'f') && (*curr < 'A' || *curr > 'F')) {
 					/* found an separator , insert value into result array */
 					put_string_into_array_len(result, base, curr - base);
 					/* change state to not in string */
 					in_str = 0;
 				}
 			} else {
-				if (*curr >= '0' && *curr <= '9') {
+				if ((*curr >= '0' && *curr <= '9') || *curr == ':' || (*curr >= 'a' && *curr <= 'f') || (*curr >= 'A' && *curr <= 'F')) {
 					/* found leading char of an IP address, move base pointer and change state */
 					base = curr;
 					in_str = 1;
@@ -328,7 +328,7 @@ static struct addrinfo *ipstr_to_sockaddr(server *srv, const char *host) {
 			"could not resolve hostname ", host, " because ", gai_strerror(result), strerror(errno));
 
 		return NULL;
-	} else if (res0 == 0) {
+	} else if (res0 == NULL) {
 		log_error_write(srv, __FILE__, __LINE__, "SSS",
 			"Problem in resolving hostname ", host, ": succeeded, but no information returned");
 	}
@@ -398,8 +398,8 @@ URIHANDLER_FUNC(mod_extforward_uri_handler) {
 	/* if the remote ip itself is not trusted, then do nothing */
 	if (IP_UNTRUSTED == is_proxy_trusted(dst_addr_str, p)) {
 		if (con->conf.log_request_handling) {
-			log_error_write(srv, __FILE__, __LINE__, "s",
-					"remote address is NOT a trusted proxy, skipping");
+			log_error_write(srv, __FILE__, __LINE__, "sss",
+					"remote address", dst_addr_str, "is NOT a trusted proxy, skipping");
 		}
 
 		return HANDLER_GO_ON;
